import numpy as np
import pytest
import jax.numpy as jnp
from chex import ArrayNumpy

from catx.network_builder import NetworkBuilder
import haiku as hk

from catx.tree import TreeParameters
from catx.type_defs import JaxObservations, Observations, Actions, Probabilities


class MLPBuilder(NetworkBuilder):
    def create_network(self, depth: int) -> hk.Module:
        return hk.nets.MLP([3] + [2 ** (depth + 1)], name=f"mlp_depth_{depth}")


@pytest.fixture
def mlp_builder() -> MLPBuilder:
    builder = MLPBuilder()

    return builder


@pytest.fixture
def tree_parameters() -> TreeParameters:
    tree_param = TreeParameters.construct(bandwidth=1 / 4, discretization_parameter=8)

    return tree_param


@pytest.fixture
def observations() -> Observations:
    obs = np.array([[0.0, 1.0], [1.0, 0.0], [1.0, 1.0], [0.0, 0.0]])

    return obs


@pytest.fixture
def jax_observations(observations: Observations) -> JaxObservations:
    obs = jnp.asarray(observations)

    return obs


@pytest.fixture
def actions() -> Actions:
    actions = np.array([0.564849, 0.57515746, 0.6175556, 0.05339688])

    return actions


@pytest.fixture
def probabilities() -> Probabilities:
    probabilities = np.array([3.85, 3.85, 3.85, 3.85], dtype=float)

    return probabilities


@pytest.fixture
def costs() -> ArrayNumpy:
    costs = np.array([0.06672329, 0.22529681, 0.03357587, 0.16503502])

    return costs


@pytest.fixture
def epsilon() -> float:
    return 0.0



from typing import Dict, Optional, TYPE_CHECKING

import haiku as hk
import numpy as np
from chex import Array
import jax.numpy as jnp

from catx.network_builder import NetworkBuilder
from catx.type_defs import Logits

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from chex import dataclass


@dataclass
class TreeParameters:
    """Dataclass for holding the tree parameters.

    Args:
        bandwidth: the bucket half width covered by action centroid.
        discretization_parameter: the number of action centroids.
        action_space: the range from which actions can be generated.
        depth: number layers in the tree.
        spaces: an array indicating the start and end range of each action centroid.
        volumes: an array indicating the bandwidth of around each action centroid.
        probabilities: h-smoothing of policy π_t (one over volumes).
    """

    bandwidth: float
    discretization_parameter: int
    action_space: Array
    depth: int
    spaces: Array
    volumes: Array
    probabilities: Array

    @classmethod
    def construct(
        cls, discretization_parameter: int, bandwidth: float
    ) -> "TreeParameters":
        """A constructor for calculating and initializing the tree parameters.

        Args:
            discretization_parameter: the number of action centroids.
            bandwidth: the bucket half width covered by action centroid.

        Returns:
            An initialized TreeParameters dataclass.
        """

        assert bandwidth > 0, "Only positive bandwidth value is admissible."
        if discretization_parameter < 2 or (
            discretization_parameter & (discretization_parameter - 1)
        ):
            raise ValueError(
                "discretization_parameter must be power of 2 number and larger than 1."
            )

        action_space = jnp.linspace(
            bandwidth, 1 - bandwidth, num=discretization_parameter
        )
        depth = int(np.log2(discretization_parameter))
        spaces = jnp.clip(
            (
                action_space.reshape((-1, 1))
                + jnp.asarray([-bandwidth, bandwidth]).reshape((1, -1))
            ),
            a_min=0,
            a_max=1,
        )
        volumes = jnp.diff(spaces).flatten()
        probabilities = 1 / volumes

        # noinspection PyArgumentList
        return cls(
            bandwidth=bandwidth,
            discretization_parameter=discretization_parameter,
            action_space=action_space,
            depth=depth,
            spaces=spaces,
            volumes=volumes,
            probabilities=probabilities,
        )


class Tree(hk.Module):
    def __init__(
        self,
        network_builder: NetworkBuilder,
        tree_params: TreeParameters,
        name: Optional[str] = None,
    ):
        """The tree as a JAX Haiku module.

        Args:
            network_builder: specify the neural network architecture for each depth in the tree.
            tree_params: object holding the tree parameters.
            name: name of the tree.
        """

        super().__init__(name)
        self.tree_params = tree_params

        self.networks = {
            depth: network_builder.create_network(depth=depth)
            for depth in range(self.tree_params.depth)
        }

    def __call__(self, obs: Array) -> Dict[int, Logits]:
        """Query the neural networks of the tree.

        Args:
            obs: the observations, i.e., batched contexts.

        Returns:
            logits: a dictionary of the networks' logits grouped pairwise.
        """

        logits = {}
        for i in range(self.tree_params.depth):
            n_leafs = 2 ** (i + 1)
            c = self.networks[i](obs).reshape(-1, n_leafs // 2, 2)
            logits[i] = c
        return logits



from unittest.mock import patch

import jax

from catx.network_builder import NetworkBuilder
import haiku as hk


@patch("catx.network_builder.NetworkBuilder.__abstractmethods__", set())
def test_network_builder() -> None:
    network_builder = NetworkBuilder()  # type: ignore
    assert "create_network" in dir(network_builder)
    assert network_builder.create_network.__isabstractmethod__  # type: ignore


def test_network_builder__create_network() -> None:
    # Sanity check
    class MLPBuilder(NetworkBuilder):
        def create_network(self, depth: int) -> hk.Module:
            return hk.nets.MLP([3] + [2 ** (depth + 1)], name=f"mlp_depth_{depth}")

    class BuilderMocker(hk.Module):
        def __init__(self, network_builder: NetworkBuilder) -> None:
            super().__init__("network_mocker_built")
            self.networks = network_builder.create_network(depth=2)

        def __call__(self) -> int:
            return 1

    def _forward() -> None:
        BuilderMocker(network_builder=builder)

    builder = MLPBuilder()

    forward = hk.transform(_forward)
    rng_key, subkey = jax.random.split(jax.random.PRNGKey(42))
    _ = forward.init(rng=subkey)



from typing import List, Tuple, Optional
import jax
import optax
import pytest
from jax import numpy as jnp
import haiku as hk
import tensorflow as tf
import numpy as np
from sklearn.datasets import fetch_openml
from catx.catx import CATX
from catx.network_builder import NetworkBuilder
from catx.type_defs import Actions, Costs, Observations


def moving_average(x: List[float], w: int) -> np.ndarray:
    return np.convolve(x, np.ones(w), "valid") / w


class OpenMLEnvironment:
    def __init__(self, dataset_id: int, batch_size: int = 5) -> None:
        self.x, self.y = fetch_openml(
            data_id=dataset_id, as_frame=False, return_X_y=True
        )
        rows_with_nan_idx = np.argwhere(np.isnan(self.x))[:, 0]
        self.x = np.delete(self.x, rows_with_nan_idx, axis=0)
        self.y = np.delete(self.y, rows_with_nan_idx, axis=0)
        self.x = self._normalize_data(self.x)
        self.y = self._normalize_data(self.y)
        self._y_mean = np.mean(self.y)
        physical_devices = tf.config.list_physical_devices("GPU")

        try:
            tf.config.experimental.set_memory_growth(physical_devices[0], True)
        except Exception:
            pass

        self.dataset = tf.data.Dataset.from_tensor_slices((self.x, self.y))
        self.dataset = self.dataset.batch(batch_size)
        self.iterator = iter(self.dataset)

    def get_new_observations(self) -> Optional[Observations]:
        try:
            x, y = self.iterator.get_next()
            self.x = x.numpy()
            self.y = y.numpy()
            return self.x
        except tf.errors.OutOfRangeError:
            return None

    def get_costs(self, actions: Actions) -> Costs:
        costs = np.abs(actions - self.y)
        return costs

    def _normalize_data(self, data: np.ndarray) -> np.ndarray:
        return (data - np.min(data, axis=0)) / (
            np.max(data, axis=0) - np.min(data, axis=0)
        )


class MLPBuilder(NetworkBuilder):
    def create_network(self, depth: int) -> hk.Module:
        return hk.nets.MLP([5, 5] + [2 ** (depth + 1)], name=f"mlp_depth_{depth}")


@pytest.mark.parametrize("dataset_id_loss", [(41540, 0.20), (42495, 0.08), (197, 0.15)])
def test_catx_convergence(dataset_id_loss: Tuple[int, float]) -> None:
    """Script to make an integration test on CATS convergence."""
    epsilon = 0.05
    dataset_id, loss_convergence_threshold = dataset_id_loss
    rng_key = jax.random.PRNGKey(42)
    catx_key, env_key = jax.random.split(rng_key, num=2)

    builder = MLPBuilder()
    optimizer = optax.adam(learning_rate=0.01)
    catx = CATX(
        rng_key=catx_key,
        network_builder=builder,
        optimizer=optimizer,
        discretization_parameter=8,
        bandwidth=1 / 8,
    )
    batch_size = 10
    environment = OpenMLEnvironment(dataset_id=dataset_id, batch_size=batch_size)
    costs_cumulative = []
    no_iterations = 1000
    for _ in range(no_iterations):
        env_key, subkey = jax.random.split(env_key)
        obs = environment.get_new_observations()
        if obs is None:
            break

        actions, probabilities = catx.sample(obs=obs, epsilon=epsilon)

        costs = environment.get_costs(actions=actions)

        catx.learn(obs, actions, probabilities, costs)

        costs_cumulative.append(jnp.mean(costs).item())

        if len(costs_cumulative) > 150:
            if np.mean(costs_cumulative[-100:]) < loss_convergence_threshold:
                break

    costs_cumulative_moving_average = moving_average(costs_cumulative, 100)
    assert costs_cumulative_moving_average[-1] < loss_convergence_threshold



from typing import List
from unittest.mock import MagicMock

import chex
import jax
import numpy as np
import pytest
import optax
from chex import ArrayNumpy, Array
from pytest_mock import MockerFixture
from jax import numpy as jnp

from catx.catx import CATX
from catx.network_builder import NetworkBuilder

from catx.type_defs import Observations, Actions, Probabilities, JaxObservations


@pytest.fixture
def catx(mlp_builder: NetworkBuilder, request: pytest.FixtureRequest = None) -> CATX:
    if not request:
        action_min = 0.0
        action_max = 1.0
    else:
        action_min = request.param[0]
        action_max = request.param[1]

    rng_key = jax.random.PRNGKey(42)
    rng_key, catx_key = jax.random.split(rng_key, num=2)

    optimizer = optax.adam(learning_rate=0.01)
    catx = CATX(
        rng_key=catx_key,
        network_builder=mlp_builder,
        optimizer=optimizer,
        discretization_parameter=4,
        bandwidth=1.5 / 4,
        action_min=action_min,
        action_max=action_max,
    )

    return catx


@pytest.fixture
def create_forward_fn_mock(mocker: MockerFixture) -> MagicMock:
    mk = mocker.patch("catx.catx.CATX._create_forward_fn")
    return mk  # type: ignore


@pytest.fixture
def create_forward_single_depth_fns_mock(mocker: MockerFixture) -> MagicMock:
    mk = mocker.patch("catx.catx.CATX._create_forward_single_depth_fns")
    mk.return_value = MagicMock(), MagicMock()

    return mk  # type: ignore


@pytest.fixture
def update_mock(mocker: MockerFixture) -> MagicMock:
    mk = mocker.patch("catx.catx.CATX._update")
    mk.return_value = MagicMock(), MagicMock(), MagicMock()
    return mk  # type: ignore


@pytest.fixture
def init_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("catx.catx.CATX._init")  # type: ignore


@pytest.fixture
def init_opt_states_mock(mocker: MockerFixture) -> MagicMock:
    mk = mocker.patch("catx.catx.CATX._init_opt_states")
    mk.return_value = MagicMock()
    return mk  # type: ignore


def test_catx__init(
    catx: CATX, jax_observations: JaxObservations, epsilon: float
) -> None:
    # Validate init initialized:
    # _params, _forward_single_depth_fns _forward_fn, _depth_params, and _opt_states attributes.
    assert not hasattr(catx, "_forward_fn")
    assert not hasattr(catx, "_params")
    assert not hasattr(catx, "_forward_single_depth_fns")
    assert not hasattr(catx, "_depth_params")
    assert not hasattr(catx, "_opt_states")
    catx._init(obs=jax_observations, epsilon=epsilon)
    assert hasattr(catx, "_forward_fn")
    assert hasattr(catx, "_params")
    assert hasattr(catx, "_forward_single_depth_fns")
    assert hasattr(catx, "_depth_params")
    assert hasattr(catx, "_opt_states")


def test_catx__sample(catx: CATX, observations: Observations, epsilon: float) -> None:
    actions, probabilities = catx.sample(obs=observations, epsilon=epsilon)

    # Validate sample return type and shape.
    assert isinstance(probabilities, ArrayNumpy)
    assert isinstance(actions, ArrayNumpy)
    chex.assert_equal_shape([actions, probabilities, observations[:, 0]])

    # Validate stochasticity in sampling.
    actions_, _ = catx.sample(obs=observations, epsilon=epsilon)
    assert not np.all(actions_ == actions)


@pytest.mark.parametrize(
    "catx",
    [(0.0, 1.0), (50.0, 100.0), (-80.0, 20.0), (-60.0, -50.0)],
    indirect=["catx"],
)
def test_catx__sample_action_range(
    catx: CATX, observations: Observations, epsilon: float
) -> None:
    actions, probabilities = catx.sample(obs=observations, epsilon=epsilon)
    assert np.all(catx._action_min <= actions)
    assert np.all(actions <= catx._action_max)


@pytest.mark.parametrize(
    "action_min, action_max", [(0.0, 0.0), (1.0, 0.0), (-5.0, -10.0)]
)
def test_catx__init_action_range_sad(
    mlp_builder: NetworkBuilder, action_min: float, action_max: float
) -> None:
    rng_key = jax.random.PRNGKey(42)
    rng_key, catx_key = jax.random.split(rng_key, num=2)
    optimizer = optax.adam(learning_rate=0.01)
    with pytest.raises(
        AssertionError,
        match=f"'action_max' must be strictly larger than 'action_min', "
        f"got: action_max={action_max} and action_min={action_min}.",
    ):
        CATX(
            rng_key=catx_key,
            network_builder=mlp_builder,
            optimizer=optimizer,
            discretization_parameter=4,
            bandwidth=1.5 / 4,
            action_min=action_min,
            action_max=action_max,
        )


def test_catx__sample_function_call_count(
    catx: CATX,
    create_forward_fn_mock: MagicMock,
    create_forward_single_depth_fns_mock: MagicMock,
    observations: Observations,
    epsilon: float,
) -> None:
    forward_fn_mock = MagicMock()
    forward_fn_mock.return_value = MagicMock(), MagicMock()
    params_mock = MagicMock()
    create_forward_fn_mock.return_value = params_mock, forward_fn_mock

    sample_call_cnt = 3
    for _ in range(sample_call_cnt):
        _, _ = catx.sample(obs=observations, epsilon=epsilon)

    assert create_forward_fn_mock.call_count == 1
    assert create_forward_single_depth_fns_mock.call_count == 1
    assert forward_fn_mock.call_count == sample_call_cnt


def test_catx__learn_calls(
    catx: CATX,
    observations: Observations,
    actions: Actions,
    probabilities: Probabilities,
    costs: ArrayNumpy,
    update_mock: MagicMock,
    init_opt_states_mock: MagicMock,
    epsilon: float,
) -> None:
    learn_call_cnt = 3
    for _ in range(learn_call_cnt):
        catx.learn(
            obs=observations, actions=actions, probabilities=probabilities, costs=costs
        )
    assert init_opt_states_mock.call_count == 1
    assert update_mock.call_count == learn_call_cnt


def test_catx__learn_param_update(
    catx: CATX,
    observations: Observations,
    actions: Actions,
    probabilities: Probabilities,
    costs: ArrayNumpy,
    epsilon: float,
) -> None:
    catx._init(jnp.asarray(observations), epsilon=epsilon)
    rng_pre = catx.rng_key
    depth_params_pre = catx._depth_params
    opt_states_pre = catx._opt_states
    params_pre = catx._params

    catx.learn(
        obs=observations, actions=actions, probabilities=probabilities, costs=costs
    )

    rng_post = catx.rng_key
    depth_params_post = catx._depth_params
    opt_states_post = catx._opt_states
    params_post = catx._params

    assert not jnp.all(rng_pre == rng_post)
    assert not jax.tree_leaves(opt_states_pre) == jax.tree_leaves(opt_states_post)
    assert_not_equal(jax.tree_leaves(params_pre), jax.tree_leaves(params_post))
    assert_not_equal(
        jax.tree_leaves(depth_params_pre), jax.tree_leaves(depth_params_post)
    )


def assert_not_equal(x1: List[Array], x2: List[Array]) -> None:
    f = lambda a, b: np.all(a == b)
    assert not np.all(list(map(f, x1, x2)))
    assert True



import functools
from typing import Callable, Tuple, Dict

import chex
import haiku as hk
import jax
import numpy as np
import optax
from chex import PRNGKey
from jax import numpy as jnp
from chex import Array
from jax.stages import Wrapped

from catx.network_builder import NetworkBuilder
from catx.tree import Tree, TreeParameters
from catx.type_defs import (
    Actions,
    Costs,
    JaxActions,
    JaxCosts,
    JaxObservations,
    JaxProbabilities,
    Logits,
    Observations,
    Probabilities,
    JaxLoss,
)


class CATX:
    """This class runs CATX action sampling and training of the tree."""

    def __init__(
        self,
        rng_key: PRNGKey,
        network_builder: NetworkBuilder,
        optimizer: optax.GradientTransformation,
        discretization_parameter: int,
        bandwidth: float,
        action_min: float = 0.0,
        action_max: float = 1.0,
    ):
        """Instantiate a CATX instance with its corresponding tree.

        Args:
            rng_key: JAX key generator.
            network_builder: specify the neural network architecture for each depth in the tree.
            optimizer: optax optimizer object.
            discretization_parameter: the number of action centroids.
            bandwidth: the bucket half width covered by action centroid.
            action_min: the lowest value of the action space.
            action_max: the highest value of the action space.
        """

        assert action_max > action_min, (
            f"'action_max' must be strictly larger than 'action_min', "
            f"got: action_max={action_max} and action_min={action_min}."
        )

        self.discretization_parameter = discretization_parameter
        self.bandwidth = bandwidth
        self.network_builder = network_builder
        self.optimizer = optimizer
        self.rng_key = rng_key
        self.tree_params = TreeParameters.construct(
            bandwidth=bandwidth, discretization_parameter=discretization_parameter
        )

        self._action_min = action_min
        self._action_max = action_max

        self.is_initialized = False

        self._forward_fn: Wrapped
        self._forward_single_depth_fns: Dict[int, Wrapped]
        self._params: hk.Params
        self._depth_params: Dict[int, hk.Params]
        self._opt_states: Dict[int, optax.OptState]

    def sample(
        self, obs: Observations, epsilon: float
    ) -> Tuple[Actions, Probabilities]:
        """Samples an action from the tree.

        Args:
            obs: the observations, i.e., batched contexts.
            epsilon: probability of selecting a random action.

        Returns:
            actions: sampled actions from the tree using epsilon-greedy.
            probabilities: the probability density value of the sampled actions.
        """

        obs = jnp.asarray(obs)

        if not self.is_initialized:
            self._init(obs=obs, epsilon=epsilon)

        self.rng_key, sample_key = jax.random.split(self.rng_key, 2)

        actions, probabilities = self._forward_fn(
            self._params, x=obs, epsilon=epsilon, rng=sample_key
        )

        return np.array(actions), np.array(probabilities)

    def learn(
        self,
        obs: Observations,
        actions: Actions,
        probabilities: Probabilities,
        costs: Costs,
    ) -> None:
        """Updates the tree:
            - updates the parameters of the depth specific neural networks.
            - copy the parameters of the depth specific neural networks to the tree neural networks.
            - update the state of the optimizers.
            - update the pseudo random key generator.

        Args:
            obs: the observations, i.e., batched contexts.
            actions: The executed actions associated with the given observations.
            probabilities: the probability density value of the actions.
            costs: The costs incurred from the executed actions.
        """

        obs = jnp.asarray(obs)

        if not self.is_initialized:
            self._init(obs=obs)

        rng_key, new_layers_params, new_opt_states = self._update(
            rng_key=self.rng_key,
            layers_params=self._depth_params,
            opt_states=self._opt_states,
            obs=obs,
            actions=actions,
            probabilities=probabilities,
            costs=costs,
        )

        self.rng_key = rng_key
        self._depth_params = new_layers_params
        self._opt_states = new_opt_states

        self._params = {
            k: v
            for layer_params in new_layers_params.values()
            for k, v in layer_params.items()
        }

    def _init(self, obs: JaxObservations, epsilon: float = 0.0) -> None:
        """Initializes the parameters of tree's neural networks,
        the forward functions, and the optimizer states.

        This functions can only be called once. It is called the first time a CATX instance is used.

        Args:
            obs: the observations, i.e., batched contexts.
        """

        if self.is_initialized:
            return

        self.is_initialized = True
        self._params, self._forward_fn = self._create_forward_fn(
            obs=obs, epsilon=epsilon
        )
        (
            self._forward_single_depth_fns,
            self._depth_params,
        ) = self._create_forward_single_depth_fns(obs)

        self._opt_states = self._init_opt_states()

    def _create_forward_fn(
        self, obs: Observations, epsilon: float
    ) -> Tuple[hk.Params, Wrapped]:
        """Creates a jitted forward function of the tree
        and initializes the parameters of tree's neural networks.

        Args:
            obs: the observations, i.e., batched contexts.

        Returns:
            _params: the parameters of the neural networks
            _forward_fn: a jitted forward function of the tree.
        """

        def _forward(
            x: JaxObservations, epsilon: float
        ) -> Tuple[JaxActions, JaxProbabilities]:
            """This forward function defines how the tree is traversed and how actions sampled:
                - All the tree logits are queried (one set of pairwise logits per tree depth).
                - The tree is traversed by following the max of the logits at each
                  tree depth until an action centroid is reached.
                - With a probability 1-epsilon an action is sampled uniformly from
                  the centroid action space and with a probability epsilon an action
                  is uniformly sampled from action space.

            Args:
                x: the observations, i.e., batched contexts.
                epsilon: probability of selecting a random action.

            Returns:
                actions: sampled actions from the tree using epsilon-greedy
                        and scaled to the environment action range.
                probabilities: the probability density value of the actions.

            """

            tree = Tree(
                network_builder=self.network_builder,
                tree_params=self.tree_params,
            )
            logits = tree(x)

            batch_size = x.shape[0]
            mask = logits[0] < jnp.max(logits[0], axis=(1, 2)).reshape(batch_size, 1, 1)

            # Traverse the tree from root to action leaf
            d = tree.tree_params.depth - 1
            for i in range(1, d):
                c = jnp.where(
                    mask.reshape(logits[i].shape[0:2] + (1,)), -jnp.inf, logits[i]
                )
                mask = c < jnp.max(c, axis=(1, 2)).reshape(batch_size, 1, 1)

            c = jnp.where(
                mask.reshape(logits[d].shape[0:2] + (1,)), -jnp.inf, logits[d]
            )
            actions_centroid = jnp.argmax(c.reshape(batch_size, -1), axis=-1)
            probabilities = tree.tree_params.probabilities[actions_centroid]

            # epsilon-greedy
            probabilities = (1 - epsilon) * probabilities + epsilon
            action_spaces = tree.tree_params.spaces[actions_centroid]

            (
                key_exploration,
                key_exploitation,
                key_sampling_exploration,
            ) = jax.random.split(hk.next_rng_key(), num=3)

            exploitation_actions = jax.random.uniform(
                key_exploitation,
                shape=actions_centroid.shape,
                minval=action_spaces[:, 0],
                maxval=action_spaces[:, 1],
            )

            exploration = jax.random.choice(
                key_exploration,
                a=jnp.asarray([False, True]),
                p=jnp.asarray([1 - epsilon, epsilon]),
                shape=actions_centroid.shape,
            )

            probabilities = jnp.where(exploration, epsilon, probabilities)

            actions = jnp.where(
                exploration,
                jax.random.uniform(
                    key_sampling_exploration, shape=actions_centroid.shape
                ),
                exploitation_actions,
            )

            # Scale sampled actions to the environment action range.
            actions = actions * (self._action_max - self._action_min) + self._action_min

            return actions, probabilities

        forward = hk.transform(_forward)

        _params = forward.init(rng=self.rng_key, x=obs, epsilon=epsilon)
        _forward_fn = jax.jit(forward.apply)

        return _params, _forward_fn

    def _create_forward_single_depth_fns(
        self, obs: Observations
    ) -> Tuple[Dict[int, Wrapped], Dict[int, hk.Params]]:
        """Creates a dictionary of jitted forward functions, one per neural networks at each tree depth
        and initializes the parameters of these neural networks.

        Args:
            obs: the observations, i.e., batched contexts.

        Returns:
            _forward_single_depth_fns: a dictionary of jitted forward functions of neural networks
                                    with tree depth as key.
            _depth_params: a dictionary of neural network parameters with tree depth as key.
        """

        def create_single_depth_function(
            depth: int,
        ) -> Callable[[JaxObservations], Logits]:
            """Creates a neural network forward function for a given depth.

            Args:
                depth: depth at which the neural network will be used.

            Returns:
                _forward: the neural network forward function.
            """

            n_leafs = 2 ** (depth + 1)

            def _forward(x: JaxObservations) -> Logits:
                """Creates a neural network forward function for a predefined depth.

                Args:
                    x: the observations, i.e., batched contexts.

                Returns:
                    the neural network forward function at the predefined depth.
                """

                tree = Tree(
                    network_builder=self.network_builder,
                    tree_params=self.tree_params,
                )
                return tree.networks[depth](x).reshape(x.shape[0], n_leafs // 2, 2)

            return _forward

        transformed_layers = {
            i: hk.transform(create_single_depth_function(i))
            for i in range(self.tree_params.depth)
        }
        _forward_single_depth_fns = {
            i: jax.jit(func.apply) for i, func in transformed_layers.items()
        }

        _depth_params = {
            i: layer.init(x=obs, rng=self.rng_key)
            for i, layer in transformed_layers.items()
        }

        return _forward_single_depth_fns, _depth_params

    def _init_opt_states(self) -> Dict[int, optax.OptState]:
        """Initializes an optimizer state for each tree depth
        using the parameters of the depth specific neural networks.

        Returns:
            a dictionary initialized optax optimizer states with tree depth as a key.
        """

        _opt_states = {
            d: self.optimizer.init(params) for d, params in self._depth_params.items()
        }

        return _opt_states

    @functools.partial(jax.jit, static_argnames=("self",))
    def _compute_smooth_costs(
        self, costs: JaxCosts, actions: JaxActions, probabilities: JaxProbabilities
    ) -> JaxCosts:
        """Computes the smooth cost for a given batch of actions.

        Args:
            costs: the costs incurred from the executed actions.
            actions: the executed actions in the tree action range [0, 1].
            probabilities: the probability density value of the actions.

        Returns:
            the smooth costs
        """

        action_space, volumes = (
            self.tree_params.action_space,
            self.tree_params.volumes,
        )
        indicator = (
            jnp.abs(actions.reshape(-1, 1) - action_space.reshape(1, -1))
            < self.bandwidth
        )

        return (
            costs.reshape(-1, 1)
            * indicator.astype(jnp.float32)
            / (volumes.reshape(1, -1) * probabilities.reshape(-1, 1))
        ).reshape(costs.shape[0], self.tree_params.discretization_parameter // 2, 2)

    @functools.partial(jax.jit, static_argnames=("self", "depth"))
    def _loss(
        self,
        layer_params: hk.Params,
        obs: JaxObservations,
        smooth_costs: JaxCosts,
        depth: int,
        mask_eq: Array,
        rng_key: chex.PRNGKey,
    ) -> JaxLoss:
        """Computes the loss function a given depth.

        Args:
            layer_params: a dictionary of neural network parameters with tree depth as key.
            obs: the observations, i.e., batched contexts.
            smooth_costs: the smooth costs.
            depth: the tree depth at which the loss will be calculated.
            mask_eq: a mask with zeros where smooth cost pairs are equal
            rng_key: JAX key generator.

        Returns:
            the sum of the batch losses.
        """

        logits = self._forward_single_depth_fns[depth](
            params=layer_params, x=obs, rng=rng_key
        )

        smooth_costs_filtered = jnp.multiply(mask_eq, smooth_costs)
        logits_filtered = jnp.multiply(mask_eq, logits)

        return jnp.sum(jax.nn.softmax(logits_filtered) * smooth_costs_filtered)

    @functools.partial(jax.jit, static_argnames=("self",))
    def _update(
        self,
        rng_key: chex.PRNGKey,
        layers_params: Dict[int, hk.Params],
        opt_states: Dict[int, optax.OptState],
        obs: JaxObservations,
        actions: JaxActions,
        probabilities: JaxProbabilities,
        costs: JaxCosts,
    ) -> Tuple[chex.PRNGKey, Dict[int, hk.Params], Dict[int, optax.OptState]]:
        """Performs the update of the neural networks.

        Args:
            rng_key: JAX key generator.
            layers_params: a dictionary of neural network parameters with tree depth as key.
            opt_states: a dictionary of optimizer states with tree depth as a key.
            obs: the observations, i.e., batched contexts.
            actions: the executed actions in the environment action range.
            probabilities: the probability density value of the actions.
            costs: the costs incurred from the executed actions.

        Returns:
            rng_key: JAX key generator.
            new_layer_params: a dictionary of the updated network parameters with tree depth as key.
            new_opt_states: a dictionary of updated optimizer states with tree depth as a key.

        """

        # Scale actions from the environment action range to the tree action range.
        actions = (actions - self._action_min) / (self._action_max - self._action_min)

        smooth_costs = self._compute_smooth_costs(
            costs=costs, actions=actions, probabilities=probabilities
        )

        new_layer_params = {}
        new_opt_states = {}

        for depth in reversed(range(self.tree_params.depth)):
            rng_key, loss_key, layer_key = jax.random.split(rng_key, num=3)
            # Create a mask with zeros where smooth cost pairs are equal
            mask_eq = jnp.logical_not(
                jnp.isclose(smooth_costs[:, :, 0], smooth_costs[:, :, 1])
            ).astype(jnp.int32)

            mask_eq = jnp.expand_dims(mask_eq, axis=-1)
            mask_eq = jnp.tile(mask_eq, reps=2)

            # Compute and apply gradient
            grads = jax.grad(self._loss)(
                layers_params[depth], obs, smooth_costs, depth, mask_eq, loss_key
            )
            updates, opt_state = self.optimizer.update(grads, opt_states[depth])
            new_layer_params[depth] = optax.apply_updates(layers_params[depth], updates)
            new_opt_states[depth] = opt_state

            # Update smooth cost for the next upper tree depth
            if depth > 0:
                # Get smooth cost from 1 step forward following the updated parameters
                logits = self._forward_single_depth_fns[depth](
                    params=new_layer_params[depth], x=obs, rng=layer_key
                )
                mask = logits < jnp.max(logits, axis=-1).reshape(
                    logits.shape[0:2] + (1,)
                )
                c = jnp.where(mask, -jnp.inf, smooth_costs)
                smooth_costs = jnp.max(c, axis=-1).reshape(
                    logits.shape[0], logits.shape[1] // 2, 2
                )

        return rng_key, new_layer_params, new_opt_states



from typing import Dict

import chex
import jax
import haiku as hk
import jax.numpy as jnp
import pytest
import numpy as np
from chex import assert_type

from catx.network_builder import NetworkBuilder
from catx.tree import TreeParameters, Tree
from catx.type_defs import JaxObservations, Logits


@pytest.mark.parametrize("bandwidth", [1.5 / 4, 1 / 8])
@pytest.mark.parametrize("discretization_parameter", [4, 8, 16])
def test_tree_parameters__creation(
    bandwidth: float, discretization_parameter: int
) -> None:
    tree_params = TreeParameters.construct(
        bandwidth=bandwidth, discretization_parameter=discretization_parameter
    )
    assert hasattr(tree_params, "bandwidth")
    assert hasattr(tree_params, "discretization_parameter")
    assert hasattr(tree_params, "action_space")
    assert hasattr(tree_params, "depth")
    assert hasattr(tree_params, "spaces")
    assert hasattr(tree_params, "volumes")
    assert hasattr(tree_params, "probabilities")

    assert isinstance(tree_params.bandwidth, float)
    assert isinstance(tree_params.discretization_parameter, int)
    assert tree_params.depth >= 1
    assert isinstance(tree_params.depth, int)
    assert_type(tree_params.action_space, float)
    assert_type(tree_params.spaces, float)
    assert_type(tree_params.volumes, float)
    assert_type(tree_params.probabilities, float)

    assert jnp.shape(tree_params.action_space)[0] == discretization_parameter

    assert int(np.log2(tree_params.discretization_parameter)) == tree_params.depth
    assert jnp.min(tree_params.spaces) >= 0
    assert jnp.max(tree_params.spaces) <= 1


@pytest.mark.parametrize("discretization_parameter", [-5, -4, 0, 1, 5, 6, 9, 12])
def test_tree_parameters__discretization_parameter(
    discretization_parameter: int,
) -> None:
    with pytest.raises(
        ValueError,
        match="discretization_parameter must be power of 2 number and larger than 1.",
    ):
        TreeParameters.construct(
            bandwidth=1 / 8, discretization_parameter=discretization_parameter
        )


@pytest.mark.parametrize("bandwidth", [-5, -1, 0])
def test_tree_parameters__bandwidth(bandwidth: float) -> None:
    with pytest.raises(
        AssertionError, match="Only positive bandwidth value is admissible."
    ):
        TreeParameters.construct(bandwidth=bandwidth, discretization_parameter=4)


def test_tree(
    mlp_builder: NetworkBuilder,
    tree_parameters: TreeParameters,
    jax_observations: JaxObservations,
) -> None:
    def _forward(x: JaxObservations) -> Dict[int, Logits]:
        tree = Tree(
            network_builder=mlp_builder,
            tree_params=tree_parameters,
        )

        output_logits = tree(x)

        # Validate the tree has as many neural networks as depth.
        assert jnp.shape(jax.tree_leaves(tree.networks))[0] == tree_parameters.depth

        return output_logits

    key = jax.random.PRNGKey(42)
    key, subkey = jax.random.split(key)

    forward = hk.transform(_forward)

    _params = forward.init(rng=subkey, x=jax_observations)

    # Validate tree initial params.
    chex.assert_tree_all_finite(_params)

    # Query the tree.
    _forward_fn = jax.jit(forward.apply)
    key, subkey = jax.random.split(key)
    logits = _forward_fn(_params, subkey, jax_observations)

    # Validate the network outputs at each depth.
    logits_shape = jax.tree_map(jnp.shape, logits)
    assert len(logits_shape) == tree_parameters.depth
    for d in range(tree_parameters.depth):
        assert logits_shape[d] == (jnp.shape(jax_observations)[0], 2**d, 2)

    chex.assert_tree_all_finite(logits)



from typing import Optional

import chex
import numpy as np

import tensorflow as tf
from sklearn.datasets import fetch_openml

from catx.type_defs import (
    Actions,
    Costs,
    Observations,
)


class OpenMLEnvironment:
    def __init__(self, dataset_id: int, batch_size: int = 5) -> None:

        self.x, self.y = fetch_openml(
            data_id=dataset_id, as_frame=False, return_X_y=True
        )

        rows_with_nan_idx = np.argwhere(np.isnan(self.x))[:, 0]
        self.x = np.delete(self.x, rows_with_nan_idx, axis=0)
        self.y = np.delete(self.y, rows_with_nan_idx, axis=0)
        self.x = self._normalize_data(self.x)
        self.y = self._normalize_data(self.y)
        self._y_mean = np.mean(self.y)
        physical_devices = tf.config.list_physical_devices("GPU")

        # Allow gpu memory growth for tensorflow.
        try:
            tf.config.experimental.set_memory_growth(physical_devices[0], True)
        except Exception:
            pass  # Invalid device or cannot modify virtual devices once initialized.

        self.dataset = tf.data.Dataset.from_tensor_slices((self.x, self.y))
        self.dataset = self.dataset.batch(batch_size)
        self.iterator = iter(self.dataset)

    def get_new_observations(self, key: chex.PRNGKey) -> Optional[Observations]:
        del key
        try:
            x, y = self.iterator.get_next()
            self.x = x.numpy()
            self.y = y.numpy()
            return self.x
        except tf.errors.OutOfRangeError:
            return None

    def get_costs(
        self, key: chex.PRNGKey, obs: Observations, actions: Actions
    ) -> Costs:
        del key
        del obs
        costs = np.abs(actions - self.y)

        return costs

    def _normalize_data(self, data: np.ndarray) -> np.ndarray:
        return (data - np.min(data, axis=0)) / (
            np.max(data, axis=0) - np.min(data, axis=0)
        )



try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from chex import Array, ArrayNumpy

Actions: TypeAlias = ArrayNumpy
JaxActions: TypeAlias = Array

Observations: TypeAlias = ArrayNumpy
JaxObservations: TypeAlias = Array

Costs: TypeAlias = ArrayNumpy
JaxCosts: TypeAlias = Array
JaxLoss: TypeAlias = Array

Probabilities: TypeAlias = ArrayNumpy
JaxProbabilities: TypeAlias = Array

Logits: TypeAlias = Array



from abc import ABC, abstractmethod

import haiku as hk


class NetworkBuilder(ABC):
    """An interface for implementing a neural network builder."""

    @abstractmethod
    def create_network(self, depth: int) -> hk.Module:
        """An abstract method for creating a Haiku neural network.

        The dimension of the network output layer should be 2**(depth+1)

        Args:
            depth: the depth at which the network will be used in the tree.

        Returns:
            a Haiku neural network module.
        """

        pass



import time
from typing import List

import haiku as hk
import jax
import optax
from jax import numpy as jnp

from catx.catx import CATX
from catx.network_builder import NetworkBuilder

import numpy as np


import matplotlib.pyplot as plt

from examples.openml_environment import OpenMLEnvironment


def moving_average(x: List[float], w: int) -> np.ndarray:
    return np.convolve(x, np.ones(w), "valid") / w


class MLPBuilder(NetworkBuilder):
    def create_network(self, depth: int) -> hk.Module:
        return hk.nets.MLP([10, 10] + [2 ** (depth + 1)], name=f"mlp_depth_{depth}")


def main() -> None:
    print('Running CATX training on the "black_friday" dataset from OpenML.')
    dataset_id = 41540  # black_friday dataset id from OpenML
    epsilon = 0.05
    start_time = time.time()

    rng_key = jax.random.PRNGKey(42)
    catx_key, env_key = jax.random.split(rng_key, num=2)

    builder = MLPBuilder()
    optimizer = optax.adam(learning_rate=0.01)
    catx = CATX(
        rng_key=catx_key,
        network_builder=builder,
        optimizer=optimizer,
        discretization_parameter=8,
        bandwidth=1 / 8,
    )

    batch_size = 10
    environment = OpenMLEnvironment(dataset_id=dataset_id, batch_size=batch_size)

    costs_cumulative = []
    for _ in range(1000):
        env_key, cost_key = jax.random.split(env_key)

        obs = environment.get_new_observations(env_key)
        if obs is None:
            break

        actions, probabilities = catx.sample(obs=obs, epsilon=epsilon)

        costs = environment.get_costs(key=cost_key, obs=obs, actions=actions)

        catx.learn(obs, actions, probabilities, costs)

        costs_cumulative.append(jnp.mean(costs).item())

    plt.plot(costs_cumulative)
    plt.plot(moving_average(costs_cumulative, 50))
    plt.title("Action costs")
    plt.show()

    print(f"CATX training took {time.time() - start_time:.1f}s")


if __name__ == "__main__":
    main()



"""Install script for setuptools."""

from typing import Any, List
from setuptools import find_packages, setup
from pathlib import Path


def read(*paths: Any, **kwargs: Any) -> str:
    """Read the contents of a text file safely."""

    with open(
        Path(__file__).parent.joinpath(*paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path: str) -> List[str]:
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


long_description = """CATX implements Contextual Bandits with Continuous Actions in JAX.
It allows implementing custom neural network architecture per tree layer.
For more information see [github repository](https://github.com/instadeepai/catx)."""

__version__ = read("catx", "VERSION")

setup(
    name="catx",
    version=__version__,
    description="Contextual Bandits with Continuous Actions in JAX",
    url="https://github.com/instadeepai/catx",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="InstaDeep",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=read_requirements("requirements.txt"),
    entry_points={"console_scripts": ["project_name = project_name.__main__:main"]},
    extras_require={
        "test": read_requirements("./requirements-test.txt"),
        "tool": read_requirements("./requirements-tool.txt"),
    },
)
